import codecs
import logging
import urllib.parse
from abc import ABC, abstractmethod
from pathlib import Path
from typing import BinaryIO, Callable, List, Optional, Tuple

import charset_normalizer
from charset_normalizer import CharsetMatch

from ggshield.core.utils import Filemode


logger = logging.getLogger(__name__)


class DecodeError(Exception):
    """
    Raised when a Scannable cannot determine the encoding of its content.

    Similar to UnicodeDecodeError, but easier to instantiate.
    """

    pass


class Scannable(ABC):
    """Base class for content that can be scanned by GGShield"""

    def __init__(self, filemode: Filemode = Filemode.FILE):
        self.filemode = filemode

    @property
    @abstractmethod
    def url(self) -> str:
        """Act as a unique identifier for the Scannable. May use custom protocols if
        required."""
        raise NotImplementedError

    @property
    @abstractmethod
    def filename(self) -> str:
        """To avoid breakage with the rest of the code base, implementations currently
        return the URL or path of the instance for now, but it should really return
        just the filename, or be removed."""
        # TODO: make this really return the filename, or remove it
        raise NotImplementedError

    @property
    @abstractmethod
    def path(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def is_longer_than(self, size: int) -> bool:
        """Return true if the length of the *decoded* content is greater than `size`.
        When possible, implementations must try to answer this without reading all
        content.
        Raise `DecodeError` if the content cannot be decoded.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def content(self) -> str:
        """Return the decoded content of the scannable"""
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} url={self.url} filemode={self.filemode}>"

    @staticmethod
    def _decode_bytes(
        raw_document: bytes, charset_match: Optional[CharsetMatch] = None
    ) -> str:
        """Low level helper function to decode bytes using `charset_match`. If
        `charset_match` is not provided, tries to determine it itself.

        Raises DecodeError if the document cannot be decoded."""
        if charset_match is None:
            charset_match = charset_normalizer.from_bytes(raw_document).best()
            if charset_match is None:
                # This means we were not able to detect the encoding
                raise DecodeError

        # Special case for utf_8 + BOM: `bytes.decode()` does not skip the BOM, so do it
        # ourselves
        if charset_match.encoding == "utf_8" and raw_document.startswith(
            codecs.BOM_UTF8
        ):
            raw_document = raw_document[len(codecs.BOM_UTF8) :]
        return raw_document.decode(charset_match.encoding, errors="replace")

    @staticmethod
    def _is_file_longer_than(fp: BinaryIO, size: int) -> Tuple[bool, Optional[str]]:
        """Helper function to implement is_longer_than() for file-based Scannable classes.

        Returns a tuple of:
        - True if file is longer than `size`, False otherwise
        - The decoded content as a string, if it has been fully read, None otherwise

        Raises DecodeError if the file cannot be decoded.
        """
        byte_content = b""
        str_content = ""

        charset_matches = charset_normalizer.from_fp(fp)
        charset_match = charset_matches.best()
        if charset_match is None:
            raise DecodeError
        fp.seek(0)
        while True:
            # Try to read more than the requested size:
            # - If the file is smaller, that changes nothing
            # - if the file is bigger, we potentially avoid a second read
            byte_chunk = fp.read(size * 2)
            if byte_chunk:
                byte_content += byte_chunk
                # Note: we decode `byte_content` and not `byte_chunk`: we can't
                # decode just the chunk because we have no way to know if it starts
                # and ends at complete code-point boundaries
                str_content = Scannable._decode_bytes(byte_content, charset_match)
                if len(str_content) > size:
                    return True, None
            else:
                # We read the whole file, keep it
                return False, str_content


class StringScannable(Scannable):
    """Implementation of Scannable for content already loaded in memory"""

    def __init__(self, url: str, content: str, filemode: Filemode = Filemode.FILE):
        super().__init__(filemode)
        self._url = url
        self._path: Optional[Path] = None
        self._content = content

    @property
    def url(self) -> str:
        return self._url

    @property
    def filename(self) -> str:
        return str(self._url)

    @property
    def path(self) -> Path:
        if self._path is None:
            result = urllib.parse.urlparse(self._url)
            self._path = Path(result.path)
        return self._path

    def is_longer_than(self, size: int) -> bool:
        return len(self._content) > size

    @property
    def content(self) -> str:
        return self._content


class Files:
    """
    Files is a list of files. Useful for directory scanning.

    TODO: Rename to something like ScannableCollection: this class is no longer limited
    to holding File instances.
    """

    def __init__(self, files: List[Scannable]):
        self._files = files

    @property
    def files(self) -> List[Scannable]:
        """The list of files owned by this instance. The same filename can appear twice,
        in case of a merge commit."""
        return self._files

    @property
    def paths(self) -> List[Path]:
        """Convenience property to list paths in the same order as files"""
        return [x.path for x in self.files]

    def __repr__(self) -> str:
        return f"<Files files={self.files}>"

    def apply_filter(self, filter_func: Callable[[Scannable], bool]) -> "Files":
        return Files([file for file in self.files if filter_func(file)])



import os
import tempfile
from typing import Any

import click
from click import UsageError
from pygitguardian import GGClient

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.git_shell import git
from ggshield.core.utils import REGEX_GIT_URL
from ggshield.scan import ScanContext, ScanMode
from ggshield.secret.repo import scan_repo_path


@click.command()
@click.argument("repository", nargs=1, type=click.STRING, required=True)
@add_secret_scan_common_options()
@click.pass_context
def repo_cmd(
    ctx: click.Context, repository: str, **kwargs: Any
) -> int:  # pragma: no cover
    """
    scan a REPOSITORY's commits at a given URL or path.

    REPOSITORY is the clone URI or the path of the repository to scan.
    Examples:

    ggshield secret scan repo git@github.com:GitGuardian/ggshield.git

    ggshield secret scan repo /repositories/ggshield
    """
    config: Config = ctx.obj["config"]
    cache: Cache = ctx.obj["cache"]
    client: GGClient = ctx.obj["client"]

    scan_context = ScanContext(
        scan_mode=ScanMode.REPO,
        command_path=ctx.command_path,
    )

    if os.path.isdir(repository):
        return scan_repo_path(
            client=client,
            cache=cache,
            output_handler=create_output_handler(ctx),
            config=config,
            scan_context=scan_context,
            repo_path=repository,
        )

    if REGEX_GIT_URL.match(repository):
        with tempfile.TemporaryDirectory() as tmpdirname:
            git(["clone", repository, tmpdirname])
            return scan_repo_path(
                client=client,
                cache=cache,
                output_handler=create_output_handler(ctx),
                config=config,
                scan_context=scan_context,
                repo_path=tmpdirname,
            )

    if any(host in repository for host in ("gitlab.com", "github.com")):
        raise UsageError(
            f"{repository} doesn't seem to be a valid git URL.\n"
            f"Did you mean {repository}.git?"
        )
    raise UsageError(f"{repository} is neither a valid path nor a git URL")



import json
from functools import partial
from typing import Any, Callable, Iterable, Iterator, List, TextIO

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.core.constants import MAX_WORKERS
from ggshield.core.errors import handle_exception
from ggshield.core.text_utils import create_progress_bar, display_info
from ggshield.scan import ScanContext, ScanMode, Scannable, StringScannable
from ggshield.secret import SecretScanCollection, SecretScanner


def generate_files_from_docsets(
    file: TextIO, verbose: bool = False
) -> Iterator[Scannable]:
    for line in file:
        obj = json.loads(line)
        documents = obj["documents"]
        for document in documents:
            if verbose:
                display_info(f"  * {document['id']}")
            yield StringScannable(document["id"], document["content"])


def create_scans_from_docset_files(
    scanner: SecretScanner,
    input_files: Iterable[TextIO],
    progress_callback: Callable[..., None],
    verbose: bool = False,
) -> List[SecretScanCollection]:
    scans: List[SecretScanCollection] = []

    for input_file in input_files:
        if verbose:
            display_info(f"- {click.format_filename(input_file.name)}")

        files = generate_files_from_docsets(input_file, verbose)
        results = scanner.scan(
            files,
            scan_threads=MAX_WORKERS,
        )
        scans.append(
            SecretScanCollection(id=input_file.name, type="docset", results=results)
        )
        progress_callback(advance=1)

    return scans


@click.command()
@click.argument("files", nargs=-1, type=click.File(), required=True)
@add_secret_scan_common_options()
@click.pass_context
def docset_cmd(
    ctx: click.Context,
    files: List[TextIO],
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    scan docset JSONL files.
    """
    config = ctx.obj["config"]
    output_handler = create_output_handler(ctx)
    try:
        with create_progress_bar(doc_type="files") as progress:

            scan_context = ScanContext(
                scan_mode=ScanMode.DOCSET,
                command_path=ctx.command_path,
            )
            scanner = SecretScanner(
                client=ctx.obj["client"],
                cache=ctx.obj["cache"],
                ignored_matches=config.secret.ignored_matches,
                scan_context=scan_context,
                ignored_detectors=config.secret.ignored_detectors,
            )
            task_scan = progress.add_task(
                "[green]Scanning content...", total=len(files)
            )
            scans = create_scans_from_docset_files(
                scanner=scanner,
                input_files=files,
                verbose=config.verbose,
                progress_callback=partial(progress.update, task_scan),
            )

        return output_handler.process_scan(
            SecretScanCollection(id=scan_context.command_id, type="docset", scans=scans)
        )
    except Exception as error:
        return handle_exception(error, config.verbose)



import os
import re
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Set

import click

from ggshield.core.file_utils import get_filepaths, is_path_binary

from .scannable import Files, Scannable


class File(Scannable):
    """Implementation of Scannable for files from the disk."""

    def __init__(self, path: str):
        super().__init__()
        self._path = Path(path)
        self._content: Optional[str] = None

    @property
    def url(self) -> str:
        return f"file://{self._path.absolute().as_posix()}"

    @property
    def filename(self) -> str:
        return str(self._path)

    @property
    def path(self) -> Path:
        return self._path

    def is_longer_than(self, size: int) -> bool:
        if self._content:
            # We already have the content, easy
            return len(self._content) > size

        if self.path.stat().st_size < size:
            # Shortcut: if the byte size is smaller than `size`, we can be sure the
            # decoded size will be smaller
            return False

        with self.path.open("rb") as fp:
            result, self._content = Scannable._is_file_longer_than(fp, size)
        return result

    @property
    def content(self) -> str:
        if self._content is None:
            with self.path.open("rb") as f:
                self._content = Scannable._decode_bytes(f.read())
        return self._content


def get_files_from_paths(
    paths: List[str],
    exclusion_regexes: Set[re.Pattern],
    recursive: bool,
    yes: bool,
    verbose: bool,
    ignore_git: bool = False,
) -> Files:
    """
    Create a scan object from files content.

    :param paths: List of file/dir paths from the command
    :param recursive: Recursive option
    :param yes: Skip confirmation option
    :param verbose: Option that displays filepaths as they are scanned
    :param ignore_git: Ignore that the folder is a git repository
    """
    filepaths = get_filepaths(
        paths, exclusion_regexes, recursive, ignore_git=ignore_git
    )
    files = list(generate_files_from_paths(filepaths, verbose))

    if verbose:
        for f in files:
            click.echo(f"- {click.format_filename(f.filename)}", err=True)

    size = len(files)
    if size > 1 and not yes:
        click.confirm(
            f"{size} files will be scanned. Do you want to continue?",
            abort=True,
            err=True,
        )

    return Files(files)


def generate_files_from_paths(
    paths: Iterable[str], verbose: bool
) -> Iterator[Scannable]:
    """Loop on filepaths and return an iterator on scannable files."""
    for path in paths:
        if os.path.isdir(path) or not os.path.exists(path):
            continue

        if is_path_binary(path):
            if verbose:
                click.echo(
                    f"ignoring binary file: {path}",
                    err=True,
                )
            continue

        yield File(path)



from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
)
from ggshield.core.errors import handle_exception
from ggshield.core.git_shell import check_git_dir
from ggshield.scan import Commit, ScanContext, ScanMode
from ggshield.secret import SecretScanCollection, SecretScanner
from ggshield.secret.output import SecretTextOutputHandler
from ggshield.secret.output.messages import remediation_message


REMEDIATION_STEPS = """  Since the secret was detected before the commit was made:
  1. replace the secret with its reference (e.g. environment variable).
  2. commit again."""

BYPASS_MESSAGE = """  - if you use the pre-commit framework:

     SKIP=ggshield git commit -m "<your message>"

  - otherwise (warning: the following command bypasses all pre-commit hooks):

     git commit -m "<your message>" --no-verify"""


@click.command()
@click.argument("precommit_args", nargs=-1, type=click.UNPROCESSED)
@add_secret_scan_common_options()
@click.pass_context
def precommit_cmd(
    ctx: click.Context, precommit_args: List[str], **kwargs: Any
) -> int:  # pragma: no cover
    """
    scan as a pre-commit git hook.
    """
    config = ctx.obj["config"]
    output_handler = SecretTextOutputHandler(
        show_secrets=config.secret.show_secrets,
        verbose=config.verbose,
        output=None,
        ignore_known_secrets=config.secret.ignore_known_secrets,
    )
    try:
        check_git_dir()

        scan_context = ScanContext(
            scan_mode=ScanMode.PRE_COMMIT,
            command_path=ctx.command_path,
        )

        commit = Commit(exclusion_regexes=ctx.obj["exclusion_regexes"])
        scanner = SecretScanner(
            client=ctx.obj["client"],
            cache=ctx.obj["cache"],
            scan_context=scan_context,
            ignored_matches=config.secret.ignored_matches,
            ignored_detectors=config.secret.ignored_detectors,
        )
        results = scanner.scan(commit.files)

        return_code = output_handler.process_scan(
            SecretScanCollection(id="cached", type="pre-commit", results=results)
        )
        if return_code:
            click.echo(
                remediation_message(
                    remediation_steps=REMEDIATION_STEPS,
                    bypass_message=BYPASS_MESSAGE,
                    rewrite_git_history=False,
                ),
                err=True,
            )
        return return_code
    except Exception as error:
        return handle_exception(error, config.verbose)



from typing import Any

import click

from ggshield.cmd.common_options import add_common_options
from ggshield.cmd.secret.ignore import ignore_cmd
from ggshield.cmd.secret.scan import scan_group


@click.group(commands={"scan": scan_group, "ignore": ignore_cmd})
@add_common_options()
def secret_group(**kwargs: Any) -> None:
    """Commands to work with secrets."""



import logging
import os
import subprocess
from functools import lru_cache
from shutil import which
from typing import List, Optional

import click
from click import UsageError

from ggshield.core.errors import UnexpectedError


COMMAND_TIMEOUT = 45

logger = logging.getLogger(__name__)


@lru_cache(None)
def _get_git_path() -> str:
    git_path = which("git")

    if git_path is None:
        raise UnexpectedError("unable to find git executable in PATH/PATHEXT")

    # lower()ing these would provide additional coverage on case-
    # insensitive filesystems but detection is problematic
    git_path = os.path.abspath(git_path)
    cwd = os.getcwd()
    path_env = [
        os.path.abspath(p) for p in os.environ.get("PATH", "").split(os.pathsep)
    ]

    # git was found - ignore git in cwd if cwd not in PATH
    if cwd == os.path.dirname(git_path) and cwd not in path_env:
        raise UnexpectedError("rejecting git executable in CWD not in PATH")

    logger.debug("Found git at %s", git_path)
    return git_path


@lru_cache(None)
def is_git_dir(wd: str) -> bool:
    try:
        git(["rev-parse", "--show-toplevel"], cwd=wd)
    except subprocess.CalledProcessError:
        return False
    return True


def check_git_dir(wd: Optional[str] = None) -> None:
    """Check if folder is git directory."""
    if wd is None:
        wd = os.getcwd()
    if not is_git_dir(wd):
        raise UsageError("Not a git directory.")


def get_git_root(wd: Optional[str] = None) -> str:
    return git(["rev-parse", "--show-toplevel"], cwd=wd)


def git(
    command: List[str],
    timeout: int = COMMAND_TIMEOUT,
    check: bool = True,
    cwd: Optional[str] = None,
) -> str:
    """Calls git with the given arguments, returns stdout as a string"""
    env = os.environ.copy()
    env["LANG"] = "C"

    try:
        logger.debug("command=%s", command)
        result = subprocess.run(
            [_get_git_path()] + command,
            check=check,
            capture_output=True,
            timeout=timeout,
            env=env,
            cwd=cwd,
        )
        if result.stderr:
            logger.debug("stderr=%s", result.stderr.decode("utf-8", errors="ignore"))
        return result.stdout.decode("utf-8", errors="ignore").rstrip()
    except subprocess.CalledProcessError as e:
        if "detected dubious ownership in repository" in e.stderr.decode(
            "utf-8", errors="ignore"
        ):
            raise UnexpectedError(
                "Git command failed because of a dubious ownership in repository.\n"
                "If you still want to run ggshield, make sure you mark "
                "the current repository as safe for git with:\n"
                "   git config --global --add safe.repository <YOUR_REPO>"
            )
        raise e
    except subprocess.TimeoutExpired:
        raise click.Abort('Command "{}" timed out'.format(" ".join(command)))


def git_ls(wd: Optional[str] = None) -> List[str]:
    cmd = ["ls-files", "--recurse-submodules"]
    return git(cmd, timeout=600, cwd=wd).split("\n")


def is_valid_git_commit_ref(ref: str) -> bool:
    """
    Check if a reference is valid and can be resolved to a commit
    """
    ref += "^{commit}"
    cmd = ["cat-file", "-e", ref]

    try:
        git(cmd)
    except subprocess.CalledProcessError:
        return False

    return True


def get_list_commit_SHA(
    commit_range: str, max_count: Optional[int] = None
) -> List[str]:
    """
    Retrieve the list of commit SHA from a range.
    :param commit_range: A range of commits (ORIGIN...HEAD)
    :param max_count: If set, limits the number of SHA returned to this amount. This
    returns the *end* of the list, so max_count=3 returns [HEAD~2, HEAD~1, HEAD].
    """

    cmd = ["rev-list", "--reverse", *commit_range.split()]
    if max_count is not None:
        cmd.extend(["--max-count", str(max_count)])
    # Makes rev-list print "bad revision" instead of telling the range is ambiguous
    cmd.append("--")

    try:
        commit_list = git(cmd).split("\n")
    except subprocess.CalledProcessError as e:
        if b"bad revision" in e.stderr and "~1.." in commit_range:
            # We got asked to list commits for A~1...B. If A~1 does not exist, but A
            # does, then return A and its descendants until B.
            a_ref, remaining = commit_range.split("~1", maxsplit=1)
            if not is_valid_git_commit_ref(f"{a_ref}~1") and is_valid_git_commit_ref(
                a_ref
            ):
                commit_range = a_ref + remaining
                return [a_ref] + get_list_commit_SHA(commit_range)
        return []

    if "" in commit_list:
        commit_list.remove("")
        # only happens when git rev-list doesn't error
        # but returns an empty range, example git rev-list HEAD...

    return commit_list



import logging
import os
from typing import Any, Mapping, Optional

import click

from ggshield.core.config.auth_config import AuthConfig
from ggshield.core.config.user_config import UserConfig
from ggshield.core.config.utils import get_attr_mapping, remove_url_trailing_slash
from ggshield.core.constants import DEFAULT_INSTANCE_URL
from ggshield.core.utils import api_to_dashboard_url, clean_url, dashboard_to_api_url


logger = logging.getLogger(__name__)


class Config:
    """
    Top-level config class. Contains an instance of UserConfig and an instance of
    AuthConfig with some magic to make it possible to access their fields directly. For
    example one can do:

    ```
    config = Config()

    # Access config.user_config.exit_zero
    print(config.exit_zero)

    # Access config.auth_config.instances
    print(config.instances)
    ```
    """

    user_config: UserConfig
    auth_config: AuthConfig
    _attr_mapping: Mapping[str, str] = get_attr_mapping(
        [(UserConfig, "user_config"), (AuthConfig, "auth_config")]
    )

    # The instance name, if ggshield is invoked with `--instance`
    _cmdline_instance_name: Optional[str]

    def __init__(self, config_path: Optional[str] = None):
        self.user_config, self._config_path = UserConfig.load(config_path=config_path)
        self.auth_config = AuthConfig.load()
        self._cmdline_instance_name = None

    def __getattr__(self, name: str) -> Any:
        try:
            subconfig_name = self._attr_mapping[name]
        except KeyError:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{name}'"
            )
        subconfig = getattr(self, subconfig_name)
        return getattr(subconfig, name)

    def __setattr__(self, key: str, value: Any) -> None:
        if key[0] == "_" or key in {"user_config", "auth_config"}:
            self.__dict__[key] = value
            return
        subconfig = getattr(self, self._attr_mapping[key])
        setattr(subconfig, key, value)

    def save(self) -> None:
        self.user_config.save(self._config_path)
        self.auth_config.save()

    @property
    def config_path(self) -> str:
        return self._config_path

    @property
    def instance_name(self) -> str:
        """
        The instance name (defaulting to URL) of the selected instance
        priority order is:
        - set from the command line (with set_cmdline_instance_name)
        - env var (in auth_config.current_instance)
        - in local user config (in user_config.dashboard_url)
        - in global user config (in user_config.dashboard_url)
        - the default instance
        """
        if self._cmdline_instance_name:
            return self._cmdline_instance_name

        try:
            url = os.environ["GITGUARDIAN_INSTANCE"]
            logger.debug("Using instance URL from $GITGUARDIAN_INSTANCE")
        except KeyError:
            pass
        else:
            return remove_url_trailing_slash(url)

        try:
            name = os.environ["GITGUARDIAN_API_URL"]
            logger.debug("Using API URL from $GITGUARDIAN_API_URL")
        except KeyError:
            pass
        else:
            return api_to_dashboard_url(name, warn=True)

        if self.user_config.instance:
            return self.user_config.instance

        return DEFAULT_INSTANCE_URL

    def set_cmdline_instance_name(self, name: str) -> None:
        """
        Override the current instance name. To be called by commands supporting the
        `--instance` option.
        """
        parsed_url = clean_url(name)
        if not parsed_url.scheme or not parsed_url.netloc:
            raise click.BadParameter(
                "Please provide a valid URL.",
                param_hint="instance",
            )
        self._cmdline_instance_name = f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def api_url(self) -> str:
        """
        The API URL to use to use the API
        It's the API URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return dashboard_to_api_url(self.instance_name)

    @property
    def dashboard_url(self) -> str:
        """
        The dashboard URL to use the dashboard
        It's the dashboard URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return self.instance_name

    @property
    def api_key(self) -> str:
        """
        The API key to use
        priority order is
        - env var
        - the api key from the selected instance
        """
        try:
            key = os.environ["GITGUARDIAN_API_KEY"]
            logger.debug("Using API key from $GITGUARDIAN_API_KEY")
        except KeyError:
            key = self.auth_config.get_instance_token(self.instance_name)
        return key

    def add_ignored_match(self, *args: Any, **kwargs: Any) -> None:
        return self.user_config.secret.add_ignored_match(*args, **kwargs)



from .iac_json_output_handler import IaCJSONOutputHandler
from .iac_output_handler import IaCOutputHandler
from .iac_text_output_handler import IaCTextOutputHandler


__all__ = [
    "IaCOutputHandler",
    "IaCJSONOutputHandler",
    "IaCTextOutputHandler",
]
